#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package operserv;

use strict;
use threads;
use threads::shared;

use constant {
	MAX_LIM => 16777215
};

*get_user_id = \&nickserv::get_user_id;
*get_user_nick = \&nickserv::get_user_nick;
*get_user_agent = \&nickserv::get_user_agent;
*is_identified = \&nickserv::is_identified;
*is_online = \&nickserv::is_online;
*notice = \&nickserv::notice;
*kill_user = \&nickserv::kill_user;

our $osnick = 'OperServ';

my $err_deny = 'Permission denied.';
my %newstypes = (
	u => 'User',
	o => 'Oper'
);

our (
	$add_akill, $del_akill, $get_all_akills, $get_expired_akills,
	$get_akill, $check_akill,

	$add_qline, $del_qline, $get_all_qlines, $get_expired_qlines,
	$get_qline, $check_qline,

	$add_logonnews, $del_logonnews, $list_logonnews, $get_logonnews,
	$consolidate_logonnews, $count_logonnews, $del_expired_logonnews,

	$add_clone_exceptname, $add_clone_exceptserver, $add_clone_exceptip,
	$del_clone_exceptname, $del_clone_exceptip,
	$list_clone_exceptname, $list_clone_exceptserver, $list_clone_exceptip,

	$flood_check, $flood_inc, $flood_expire,

	$get_session_list
);

sub init() {
	$add_akill = $services::dbh->prepare("INSERT INTO akill SET setter=?, mask=?, reason=?, time=?, expire=?");
	$del_akill = $services::dbh->prepare("DELETE FROM akill WHERE mask=?");
	$get_all_akills = $services::dbh->prepare("SELECT setter, mask, reason, time, expire FROM akill ORDER BY time ASC");
	$get_akill = $services::dbh->prepare("SELECT setter, mask, reason, time, expire FROM akill WHERE mask=?");
	$check_akill = $services::dbh->prepare("SELECT 1 FROM akill WHERE mask=?");

	$get_expired_akills = $services::dbh->prepare("SELECT setter, mask, reason, time, expire FROM akill WHERE expire < UNIX_TIMESTAMP() AND expire!=0");

	$add_qline = $services::dbh->prepare("INSERT INTO qline SET setter=?, mask=?, reason=?, time=?, expire=?");
	$del_qline = $services::dbh->prepare("DELETE FROM qline WHERE mask=?");
	$get_all_qlines = $services::dbh->prepare("SELECT setter, mask, reason, time, expire FROM qline ORDER BY time ASC");
	$get_qline = $services::dbh->prepare("SELECT setter, mask, reason, time, expire FROM qline WHERE mask=?");
	$check_qline = $services::dbh->prepare("SELECT 1 FROM qline WHERE mask=?");

	$get_expired_qlines = $services::dbh->prepare("SELECT mask FROM qline WHERE expire < UNIX_TIMESTAMP() AND expire!=0");

	$add_logonnews = $services::dbh->prepare("INSERT INTO logonnews SET setter=?, expire=?, type=?, id=?, msg=?, time=UNIX_TIMESTAMP()");
	$del_logonnews = $services::dbh->prepare("DELETE FROM logonnews WHERE type=? AND id=?");
	$list_logonnews = $services::dbh->prepare("SELECT setter, time, expire, id, msg FROM logonnews WHERE type=? ORDER BY id ASC");
	$get_logonnews = $services::dbh->prepare("SELECT setter, time, msg FROM logonnews WHERE type=? ORDER BY id ASC");
	$consolidate_logonnews = $services::dbh->prepare("UPDATE logonnews SET id=id-1 WHERE type=? AND id>?");
	$count_logonnews = $services::dbh->prepare("SELECT COUNT(*) FROM logonnews WHERE type=?");
	$del_expired_logonnews = $services::dbh->prepare("DELETE FROM logonnews WHERE expire < UNIX_TIMESTAMP() AND expire!=0");
	
	$add_clone_exceptname = $services::dbh->prepare("REPLACE INTO sesexname SET host=?, serv=0, adder=?, lim=?");
	$add_clone_exceptserver = $services::dbh->prepare("REPLACE INTO sesexname SET host=?, serv=1, adder=?, lim=?");
	$add_clone_exceptip = $services::dbh->prepare("REPLACE INTO sesexip SET ip=INET_ATON(?), mask=?, adder=?, lim=?");

	$del_clone_exceptname = $services::dbh->prepare("DELETE FROM sesexname WHERE host=?");
	$del_clone_exceptip = $services::dbh->prepare("DELETE FROM sesexip WHERE ip=INET_ATON(?)");

	$list_clone_exceptname = $services::dbh->prepare("SELECT host, adder, lim FROM sesexname WHERE serv=0 ORDER BY host ASC");
	$list_clone_exceptserver = $services::dbh->prepare("SELECT host, adder, lim FROM sesexname WHERE serv=1 ORDER BY host ASC");
	$list_clone_exceptip = $services::dbh->prepare("SELECT INET_NTOA(ip), mask, adder, lim FROM sesexip ORDER BY ip ASC");

	$flood_check = $services::dbh->prepare("SELECT flood FROM user WHERE id=?");
	$flood_inc = $services::dbh->prepare("UPDATE user SET flood = flood + ? WHERE id=?");
	$flood_expire = $services::dbh->prepare("UPDATE user SET flood = flood >> 1"); # shift is faster than mul

	$get_session_list = $services::dbh->prepare("SELECT host, COUNT(*) AS c FROM user WHERE online=1 GROUP BY host HAVING c >= ?");
}

sub dispatch($$$) {
	my ($src, $dst, $msg) = @_;
	$msg =~ s/^\s+//;
	my @args = split(/\s+/, $msg);
	my $cmd = shift @args;

	my $user = { NICK => $src, AGENT=> $osnick };

	services::ulog($osnick, log::INFO(), "cmd: [$msg]", $user);

	return if flood_check($user);
	unless(adminserv::is_ircop($user)) {
		notice($user, $err_deny);
		net::globops($osnick, "\002$src\002 failed access to $osnick $msg");
		return;
	}

	if ($cmd =~ /^fjoin$/i) 	{ os_fjoin($user, @args); }
	elsif ($cmd =~ /^fpart$/i) 	{ os_fpart($user, @args); }
	elsif ($cmd =~ /^unidentify$/i)	{ os_unidentify($user, @args); }
	elsif ($cmd =~ /^qline$/i) {
		my $cmd2 = shift @args;

		if($cmd2 =~ /^add$/i) {
			if(@args >= 3 and $args[0] =~ /^\+/) {
				@args = split(/\s+/, $msg, 5);
				
				os_qline_add($user, $args[2], $args[3], $args[4]);
			}
			elsif(@args >= 2) {
				@args = split(/\s+/, $msg, 4);
				
				os_qline_add($user, 0, $args[2], $args[3]);
			}
			else {
				notice($user, 'Syntax: QLINE ADD [+expiry] <mask> <reason>');
			}
		}
		elsif($cmd2 =~ /^del$/i) {
			if(@args == 1) {
				os_qline_del($user, $args[0]);
			}
			else {
				notice($user, 'Syntax: QLINE DEL <mask>');
			}
		}
		elsif($cmd2 =~ /^list$/i) {
			if(@args == 0) {
				os_qline_list($user);
			}
			else {
				notice($user, 'Syntax: QLINE LIST');
			}
		}
	}	
	elsif ($cmd =~ /^jupe$/i) {
		if(@args >= 2) {
			os_jupe($user, shift @args, join(' ', @args));
		}
		else {
			notice($user, 'Syntax: JUPE <server> <reason>');
		}
	}
	elsif ($cmd =~ /^uinfo$/i)	{ os_uinfo($user, @args); }
	elsif ($cmd =~ /^svsnick$/i)	{ os_svsnick($user, $args[0], $args[1]); }
	elsif ($cmd =~ /^gnick$/i)	{ os_gnick($user, @args); }
	elsif ($cmd =~ /^help$/i)	{ help::sendhelp($osnick, $src, 'operserv', @args) }
	elsif ($cmd =~ /^(staff|listadm)$/i)	{ adminserv::as_staff($user) }
	elsif ($cmd =~ /^logonnews$/i) {
		my $cmd2 = shift @args;

		if($cmd2 =~ /^add$/i) {
			if(@args >= 3 and $args[1] =~ /^\+/) {
				@args = split(/\s+/, $msg, 5);

				os_logonnews_add($user, $args[2], $args[3], $args[4]);
			}
			elsif(@args >= 2) {
				@args = split(/\s+/, $msg, 4);

				os_logonnews_add($user, $args[2], 0, $args[3]);
			}
			else {
				notice($user, 'Syntax: LOGONNEWS ADD <type> [+expiry] <reason>');
			}
		}
		elsif($cmd2 =~ /^del$/i) {
			if(@args == 2) {
				os_logonnews_del($user, $args[0], $args[1]);
			}
			else {
				notice($user, 'Syntax: LOGONNEWS DEL <type> <id>');
			}
		}
		elsif($cmd2 =~ /^list$/i) {
			if(@args == 1) {
				os_logonnews_list($user, $args[0]);
			}
			else {
				notice($user, 'Syntax: LOGONNEWS LIST <type>');
			}
		}
		else {
			notice($user, 'Syntax: LOGONNEWS <LIST|ADD|DEL> <type>');
		}
	}
	elsif($cmd =~ /^except(ion)?$/i) {
		my $cmd2 = shift @args;
		if($cmd2 =~ /^server$/i) {
			my $cmd3 = shift @args;
			if($cmd3 =~ /^a(dd)?$/) {
				if(@args == 2) {
					os_except_server_add($user, $args[0], $args[1]);
				}
				else {
					notice($user, 'Syntax EXCEPT SERVER ADD <hostname> <limit>');
				}
			}
			elsif($cmd =~ /^d(el)?$/) {
				if(@args == 1) {
					os_except_server_del($user, $args[0]);
				}
				else {
					notice($user, 'Syntax EXCEPT SERVER DEL <hostname>');
				}
			}
			else {
				notice($user, 'Syntax EXCEPT SERVER <ADD|DEL>');
			}
		}
		elsif($cmd2 =~ /^h(ostname)?$/i) {
			my $cmd3 = shift @args;
			if($cmd3 =~ /^a(dd)?$/) {
				if(@args == 2) {
					os_except_hostname_add($user, $args[0], $args[1]);
				}
				else {
					notice($user, 'Syntax EXCEPT HOSTNAME ADD <hostname> <limit>');
				}
			}
			elsif($cmd3 =~ /^d(el)?$/) {
				if(@args == 1) {
					os_except_hostname_del($user, $args[0]);
				}
				else {
					notice($user, 'Syntax EXCEPT HOSTNAME DEL <hostname>');
				}
			}
			else {
				notice($user, 'Syntax EXCEPT HOSTNAME <ADD|DEL>');
			}
		}
		elsif($cmd2 =~ /^i(p)?$/i) {
			my $cmd3 = shift @args;
			if($cmd3 =~ /^a(dd)?$/) {
				if(@args == 2) {
					os_except_IP_add($user, $args[0], $args[1]);
				}
				else {
					notice($user, 'Syntax EXCEPT IP ADD <IP/mask> <limit>');
				}
			}
			elsif($cmd3 =~ /^d(el)?$/) {
				if(@args == 1) {
					os_except_IP_del($user, $args[0]);
				}
				else {
					notice($user, 'Syntax EXCEPT IP DEL <IP>');
				}
			}
			else {
				notice($user, 'Syntax EXCEPT IP <ADD|DEL>');
			}
		}
		elsif($cmd2 =~ /^l(ist)?$/i) {
			if(@args == 0) {
				os_except_list($user);
			}
			else {
				notice($user, 'Syntax EXCEPT LIST');
			}
		}
		else {
			notice($user, 'Syntax: EXCEPT <SERVER|HOSTNAME|IP|LIST>');
		}
	}
	elsif($cmd =~ /^session$/i) {
		if(@args == 1) {
			os_session_list($user, $args[0]);
		} else {
			notice($user, 'Syntax SESSION <lim>');
		}
	}
	elsif($cmd =~ /^chankill$/i) {
		if(@args >= 2) {
			@args = split(/\s+/, $msg, 2);
			os_chankill($user, @args);
		} else {
			notice($user, 'Syntax: CHANKILL <#chan> <reason>');
		}
	}
	elsif ($cmd =~ /^rehash$/i) {
		if(@args <= 1) {
			os_rehash($user, @args);
		}
		else {
			notice($user, 'Syntax: REHASH [type]');
		}
	}
	elsif ($cmd =~ /^loners$/i) {
		os_loners($user, @args);
	}
	elsif($cmd =~ /^svskill$/i) {
		if(@args >= 2) {
			os_svskill($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax SVSKILL <target> <reason here>');
		}
	}

	else { notice($user, "Unknown command."); }
}

sub os_fjoin($$@) {
	my ($user, $target, @chans) = @_;
	if ((!$target or !@chans) or !($chans[0] =~ /^#/)) {
		notice($user, "Syntax: /OS FJOIN <nick> <#channel1> [#channel2]");
	}
	unless (is_online($target)) {
		notice($user, "\002$target\002 is not online");
		return;
	}
    
	if (!adminserv::can_do($user, 'FJOIN')) {
		notice($user, "You don't have the right access");
		return $event::SUCCESS;
	}
	net::svsjoin($osnick, $target, @chans);
}

sub os_fpart($$@) {
	my ($user, $target, @params) = @_;
	if ((!$target or !@params) or !($params[0] =~ /^#/)) {
		notice($user, "Syntax: /OS FPART <nick> <#channel1> [#channel2] [reason]");
	}
	unless (is_online($target)) {
		notice($user, "\002$target\002 is not online");
		return;
	}

	if (!adminserv::can_do($user, 'FJOIN')) {
		notice($user, "You don't have the right access");
		return $event::SUCCESS;
	}
	
	my ($reason, @chans);
	while ($params[0] =~ /^#/) {
		push @chans, shift @params;
	}
	$reason = join(' ', @params) if @params;
	
	net::svspart($osnick, $target, $reason, @chans);
}

sub os_qline_add($$$$) {
	my ($user, $expiry, $mask, $reason) = @_;
	
	chk_auth($user, 'QLINE') or return;
	
	$expiry = misc::parse_time($expiry);
	if($expiry) { $expiry += time() }
	else { $expiry = 0 }
	
	$check_qline->execute($mask);
	if ($check_qline->fetchrow_array) {
		notice($user, "$mask is already qlined");
		return $event::SUCCESS;
	} else {
		my $src = get_user_nick($user);
		$add_qline->execute($src, $mask, $reason, time(), $expiry);
		net::sqline($mask, $reason);
		notice($user, "$mask is now Q:lined");
	}
}

sub os_qline_del($$) {
	my($user, $mask) = @_;
	
	chk_auth($user, 'QLINE') or return;
	
	$check_qline->execute($mask);
	if($check_qline->fetchrow_array) {
		$del_qline->execute($mask);
		net::unsqline($mask);
		notice($user, "$mask unqlined");
	} else {
		notice($user, "$mask is not qlined");
	}
}

sub os_qline_list($) {
	my ($user) = @_;
	my (@reply);

	chk_auth($user, 'QLINE') or return;

	push @reply, 'Q:line list:';
	
	$get_all_qlines->execute();
	my $i;
	while (my ($setter, $mask, $reason, $time, $expiry) = $get_all_qlines->fetchrow_array) {
		$i++;
		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = misc::gmtime2($time);
		my $akill_entry1 = "  $i) \002$mask\002  $reason";
		my $akill_entry2 = "    set by $setter on ".scalar(misc::gmtime2($time)).'; ';
		if($expiry) {
			my ($weeks, $days, $hours, $minutes, $seconds) = misc::split_time($expiry-time());
			$akill_entry2 .= "Expires in ".($weeks?"$weeks weeks ":'').
				($days?"$days days ":'').
				($hours?"$hours hours ":'').
				($minutes?"$minutes minutes ":'');
		}
		else {
			$akill_entry2 .= "Does not expire.";
		}
		push @reply, $akill_entry1; push @reply, $akill_entry2;
	}
	$get_all_qlines->finish();
	push @reply, ' --';

	notice($user, @reply) if @reply;
}

sub os_jupe($$$) {
	# introduces fake server to network.
	my ($user, $server, $reason) = @_;

	unless (adminserv::is_svsop($user, adminserv::S_ROOT())) {
	    notice($user, $err_deny);
	    return $event::SUCCESS;
	}

	net::jupe_server($server, "Juped by ".get_user_nick($user).": $reason");
	return $event::SUCCESS;
}

sub os_unidentify($$) {
	my ($user, $tnick) = @_;
	
	my $tuser = { NICK => $tnick };
	my $tuid;
	
	unless ($tuid = get_user_id($tuser)) {
		notice($user, "\002$tnick\002 is not online");
	}
	unless (adminserv::can_do($user, 'SERVOP')) {
		notice($user, $err_deny);
	}
	$nickserv::logout->execute($tuid);
	notice($user, "$tnick logged out from all nick identifies");
}

sub os_uinfo($@) {
	my ($user, @targets) = @_;

	my @userlist;
	foreach my $target (@targets) {
		my @data;
		my $tuser = { NICK => $target };
		my $tuid = get_user_id($tuser);
		unless ($tuid) {
			notice($user, "\002$target\002 is not online");
			next;
		}
		push @userlist, $tuser;
	}

	notice($user, get_uinfo(@userlist));
	return $event::SUCCESS;
}

sub os_svsnick($$$) {
    my ($user, $curnick, $newnick) = @_;
    my $tuser = { NICK => $curnick };
    
    if(!adminserv::is_svsop($user, adminserv::S_ROOT())) {
	notice($user, $err_deny);
	return $event::SUCCESS;
    }
    if ((!$curnick) or (!$newnick)) {
	notice($user, "Syntax: SVSNICK <curnick> <newnick>");
	return $event::SUCCESS;
    }
    if (!is_online($tuser)) {
	notice($user, $curnick.' is not online.');
	return $event::SUCCESS;
    }
    if (nickserv::is_online($newnick)) {
	notice($user, $newnick.' already exists.');
	return $event::SUCCESS;
    }
    nickserv::enforcer_quit($newnick);
    net::svsnick($osnick, $curnick, $newnick);
    notice($user, $curnick.' changed to '.$newnick);
    return $event::SUCCESS;
}

sub os_gnick($@) {
	my ($user, @targets) = @_;
    
	if(!adminserv::can_do($user, 'QLINE')) {
		notice($user, $err_deny);
		return $event::SUCCESS;
	}
	if (@targets == 0) {
		notice($user, "Syntax: GNICK <nick>");
		return $event::SUCCESS;
	}
	foreach my $target (@targets) {
		if (!is_online($target)) {
			notice($user, $target.' is not online.');
			next;
		}
		my $newnick = nickserv::collide($target);
		notice($user, $target.' changed to '.$newnick);
	}
	return $event::SUCCESS;
}

sub os_logonnews_pre($$) {
	my ($user, $type) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_ADMIN())) {
		notice($user, $err_deny);
		return undef;
	}

	return 'u' if($type =~ /^(user)|(u)$/i);
	return 'o' if($type =~ /^(oper)|(o)$/i);
	notice($user, 'invalid LOGONNEWS <type>');
	return undef;
}

sub os_logonnews_add($$$) {
	my ($user, $type, $expiry, $msg) = @_;

	return unless ($type = os_logonnews_pre($user, $type));

	my $mlength = length($msg);
	if($mlength >= 350) {
		notice($user, 'Message is too long by '. $mlength-350 .' character(s). Maximum length is 350 chars');
		return;
	}

	if($expiry) {
		$expiry = misc::parse_time($expiry);
	}
	else {
		$expiry = 0;
	}

	my $src = get_user_nick($user);
	$count_logonnews->execute($type);
	my $count = $count_logonnews->fetchrow_array;

	$add_logonnews->execute($src, $expiry ? time()+$expiry : 0, $type, ++$count, $msg);

	notice($user, "Added new $newstypes{$type} News #\002$count\002");
}

sub os_logonnews_del($$$) {
	my ($user, $type, $id) = @_;

	return unless ($type = os_logonnews_pre($user, $type));

	my $ret = $del_logonnews->execute($type, $id);

	if ($ret == 1) {
		notice($user, "News Item $newstypes{$type} News #\002$id\002 deleted");
		$consolidate_logonnews->execute($type, $id);
	}
	else {
		notice($user, "Delete of $newstypes{$type} News #\002$id\002 failed.",
			"$newstypes{$type} #\002$id\002 does not exist?");
	}
}

sub os_logonnews_list($$) {
	my ($user, $type) = @_;

	return unless ($type = os_logonnews_pre($user, $type));

	my @reply;
	push @reply, "\002$newstypes{$type}\002 News";

	$list_logonnews->execute($type);
	push @reply, "There is no $newstypes{$type} News"
		unless($list_logonnews->rows);
	while(my ($adder, $time, $expiry, $id, $msg) = $list_logonnews->fetchrow_array) {
		my ($weeks, $days, $hours, $minutes, $seconds) = misc::split_time($expiry-time());
		my $expire_string = ($expiry?"Expires in ".($weeks?"$weeks weeks ":'').
			($days?"$days days ":'').
			($hours?"$hours hours ":'').
			($minutes?"$minutes minutes ":'')
			:'Does not expire');
		push @reply, "$id\) $msg";
		push @reply, fmt::columnar2(['added: '.misc::gmtime2($time), $expire_string, "added by: $adder"]);
	}
	$list_logonnews->finish();
	notice($user, @reply);
}

sub os_except_pre($) {
	my ($user) = @_;

	if (adminserv::is_svsop($user, adminserv::S_ADMIN()) ) {
		return 1;
	}
	else {
		notice($user, $err_deny);
		return 0;
	}
}

sub os_except_hostname_add($$$) {
	my ($user, $hostname, $limit) = @_;

	os_except_pre($user) or return 0;

	if ($hostname =~ m/\@/ or not $hostname =~ /\./) {
		notice($user, 'Invalid hostmask.', 'A clone exception hostmask is the HOST portion only, no ident',
			'and must contain at least one dot \'.\'');
		return;
	}

	$limit = MAX_LIM() unless $limit;

	my $src = get_user_nick($user);
	my $hostmask = $hostname;
	$hostmask =~ s/\*/\%/g;
	$add_clone_exceptname->execute($hostmask, $src, $limit);
	notice($user, "Clone exception for host \002$hostname\002 added.");
}

sub os_except_server_add($$$) {
	my ($user, $hostname, $limit) = @_;

	os_except_pre($user) or return 0;

	if ($hostname =~ m/\@/ or not $hostname =~ /\./) {
		notice($user, 'Invalid hostmask.', 'A clone exception servername has no ident',
			'and must contain at least one dot \'.\'');
		return;
	}

	$limit = MAX_LIM() unless $limit;

	my $src = get_user_nick($user);
	my $hostmask = $hostname;
	$hostmask =~ s/\*/\%/g;
	$add_clone_exceptserver->execute($hostmask, $src, $limit);
	notice($user, "Clone exception for server \002$hostname\002 added.");
}

sub os_except_IP_add($$$$) {
	my ($user, $IP, $limit) = @_;

	os_except_pre($user) or return 0;

	my $mask;
	($IP, $mask) = split(/\//, $IP);
	$mask = 32 unless $mask;
	if ($IP =~ m/\@/ or not $IP =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/) {
		notice($user, 'Invalid hostmask.', 'A clone exception IP has no ident',
			'and must be a valid IP address with 4 octets (example: 1.2.3.4)');
		return;
	}

	$limit = MAX_LIM() unless $limit;

	my $src = get_user_nick($user);
	$add_clone_exceptip->execute($IP, $mask, $src, $limit);
	notice($user, "IP clone exception \002$IP\/$mask\002 added.");
}

sub os_except_hostname_del($$) {
	my ($user, $hostname) = @_;

	os_except_pre($user) or return 0;
	
	my $hostmask = $hostname;
	$hostmask =~ s/\*/\%/g;
	my $ret = $del_clone_exceptname->execute($hostmask);
	net::notice($osnick, $config::conf{diag}, "hostname: $hostname; hostmask: $hostmask");
	
	if($ret == 1) {
		notice($user, "\002$hostname\002 successfully deleted from the hostname exception list");
	}
	else {
		notice($user, "Deletion of \002$hostname\002 \037failed\037. \002$hostname\002 entry does not exist?");
	}
}

sub os_except_server_del($$) {
	my ($user, $hostname) = @_;

	os_except_pre($user) or return 0;
	
	my $hostmask = $hostname;
	$hostmask =~ s/\*/\%/g;
	my $ret = $del_clone_exceptname->execute($hostmask);
	
	if($ret == 1) {
		notice($user, "\002$hostname\002 successfully deleted from the server exception list");
	}
	else {
		notice($user, "Deletion of \002$hostname\002 \037failed\037. \002$hostname\002 entry does not exist?");
	}
}

sub os_except_IP_del($$$) {
	my ($user, $IP) = @_;

	os_except_pre($user) or return 0;
	
	my ($IP, $mask) = split(/\//, $IP);
	$mask = 32 unless $mask;
	my $ret = $del_clone_exceptip->execute($IP);
	
	if($ret == 1) {
		notice($user, "\002$IP/$mask\002 successfully deleted from the IP exception list");
	}
	else {
		notice($user, "Deletion of \002$IP/$mask\002 \037failed\037. \002$IP/$mask\002 entry does not exist?");
	}
}

sub os_except_list($) {
	my ($user) = @_;
	my @data;

	$list_clone_exceptserver->execute();
	while(my ($host, $adder, $lim) = $list_clone_exceptserver->fetchrow_array) {
		$host =~ s/\%/\*/g;
		push @data, ['Server:', $host, $lim!=MAX_LIM()?$lim:'unlimited', "($adder)"];
	}

	$list_clone_exceptname->execute();
	while(my ($host, $adder, $lim) = $list_clone_exceptname->fetchrow_array) {
		$host =~ s/\%/\*/g;
		push @data, ['Host:', $host, $lim!=MAX_LIM()?$lim:'unlimited', "($adder)"];
	}
	
	$list_clone_exceptip->execute();
	while(my ($ip, $mask, $adder, $lim) = $list_clone_exceptip->fetchrow_array) {
		push @data, ['IP:', "$ip/$mask", $lim!=MAX_LIM()?$lim:'unlimited', "($adder)"];
	}
	
	notice($user,
		"Clone exception list:",
		fmt::columnar(@data)
	);
}

sub os_session_list($) {
	my ($user, $lim) = @_;

	unless($lim > 1) {
		notice($user, "Please specify a number greater than 1.");
		return;
	}

	$get_session_list->execute($lim);
	my $data = $get_session_list->fetchall_arrayref;

	notice($user,
		"Hosts with at least $lim sessions:",
		fmt::columnar(@$data)
	);
}

sub os_chankill($$$) {
	my ($user, $cn, $reason) = @_;
	my $src = get_user_nick($user);

	chanserv::chan_kill({ CHAN=> $cn }, "$reason ($src - ".misc::gmtime2(time()).")");
}

sub os_rehash($;$) {
	my ($user, $type) = @_;

	unless (adminserv::is_svsop($user, adminserv::S_ROOT())) {
	    notice($user, $err_deny);
	    return $event::SUCCESS;
	}

	net::rehash_all_servers($type);
	return $event::SUCCESS;
}

sub os_loners($@) {
	my ($user, @args) = @_;
	my $cmd = shift @args;
	my $noid;
	if (defined($args[0]) and $args[0] =~ /(not?id|noidentify)/) {
		$noid = 1;
		shift @args;
	}

	if(!$cmd or $cmd =~ /^(list|not?id|noidentify)$/i) {
		if (defined($cmd) and $cmd =~ /(not?id|noidentify)/) {
			$noid = 1;
		}
		my @reply;
		foreach my $tuser (chanserv::get_users_nochans($noid)) {
			push @reply, get_user_nick($tuser);
		}
		notice($user, "Users in zero channels", @reply);
	}
	elsif($cmd =~ /^uinfo$/i) {
		notice($user, get_uinfo(chanserv::get_users_nochans($noid)));
	}
	elsif($cmd =~ /^kill$/i) {
		unless(adminserv::is_svsop($user, adminserv::S_OPER())) {
			notice($user, $err_deny);
			return;
		}
		foreach my $tuser (chanserv::get_users_nochans($noid)) {
			$tuser->{AGENT} = $osnick;
			nickserv::kill_user($tuser,
				"Killed by \002".get_user_nick($user)."\002".
				(@args ? ": ".join(' ', @args) : '')
			);
		}
	}
	elsif($cmd =~ /^kline$/i) {
		unless(adminserv::is_svsop($user, adminserv::S_OPER())) {
			notice($user, $err_deny);
			return;
		}
		foreach my $tuser (chanserv::get_users_nochans($noid)) {
			$tuser->{AGENT} = $osnick;
			nickserv::kline_user($tuser, $services::conf{chankilltime},
				"K:Lined by \002".get_user_nick($user)."\002".
				(@args ? ": ".join(' ', @args) : '')
			);
		}
	}
	elsif($cmd =~ /^(msg|message|notice)$/i) {
		notice($user, "Must have message to send") unless(@args);
		foreach my $tuser (chanserv::get_users_nochans($noid)) {
			$tuser->{AGENT} = $osnick;
			notice($tuser, 
				"Automated message from \002".get_user_nick($user),
				join(' ', @args)
			);
		}
	}
	elsif($cmd =~ /^fjoin$/i) {
		unless(adminserv::can_do($user, 'FJOIN')) {
			notice($user, $err_deny);
			return;
		}

		if ($args[0] !~ /^#/) {
			notice($user, "\002".$args[0]."\002 is not a valid channel name");
			return;
		}

		foreach my $tuser (chanserv::get_users_nochans($noid)) {
			$tuser->{AGENT} = $osnick;
			net::svsjoin($osnick, get_user_nick($tuser), $args[0]);
		}
	}
	else {
		notice($user, "Unknown LONERS command: $cmd",
			'Syntax: OS LONERS [LIST|UINFO|MSG|FJOIN|KILL|KLINE] [NOID] [msg/reason]');
	}
}

sub os_svskill($$$) {
	my ($user, $target, $reason) = @_;
	my $tuser = { NICK => $target };
	
	if(!adminserv::is_svsop($user, adminserv::S_ROOT())) {
		notice($user, $err_deny);
		return $event::SUCCESS;
	}

	if (!is_online($tuser)) {
		notice($user, $target.' is not online.');
		return $event::SUCCESS;
	}

	net::svskill($osnick, $target, $reason);

	return $event::SUCCESS;
}

### MISCELLANEA ###

sub do_news($$) {
	my ($nick, $type) = @_;

	my ($banner, @reply);

	if ($type eq 'u') {
		$banner = "\002Logon News\002";
	}
	elsif ($type eq 'o') {
		$banner = "\002Oper News\002";
	}
	$get_logonnews->execute($type);
	while(my ($adder, $time, $msg) = $get_logonnews->fetchrow_array) {
		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($time);
		$year += 1900;
		push @reply, "[$banner ".$misc::months[$mon]." $mday $year] $msg";
	}
	$get_logonnews->finish();
	net::notice($config::conf{local}, $nick, @reply);
}

sub chk_auth($$) {
	my ($user, $perm) = @_;
	
	if(adminserv::can_do($user, $perm)) {
		return 1;
	}
	
	notice($user, $err_deny);
	return 0;
}

sub expire(;$) {
	event::addtimer('OperServ Expire', 60, __PACKAGE__, 'operserv::expire');

	$get_expired_qlines->execute();
	while (my ($mask) = $get_expired_qlines->fetchrow_array() ) {
		net::unsqline($mask);
		$del_qline->execute($mask);
	}
	$get_expired_qlines->finish();
	
	#don't run this code yet.
#	$get_expired_akills->execute();
#	while (my ($mask) = $get_expired_akills->fetchrow_array() ) {
#		($ident, $host) = split('@', $mask);
#		net::unkline($osnick, $ident, $host);
#		$del_akill->execute($mask);
#	}
#	$get_expired_akills->finish();

	$del_expired_logonnews->execute();
}

sub flood_expire(;$) {
	event::addtimer('flood_expire', 10, __PACKAGE__, 'operserv::flood_expire');
	$flood_expire->execute();
	$flood_expire->finish();
}

sub flood_check($;$) {
	my ($user, $amt) = @_;
	$amt = 1 unless defined($amt);

	$flood_inc->execute($amt, get_user_id($user)) unless adminserv::is_svsop($user, adminserv::S_HELP());
	
	my $flev = get_flood_level($user);
	if($flev > 8) {
		kill_user($user, "Flooding services.");
		return 1;
	}
	elsif($flev > 6) {
		notice($user, "You are flooding services.") if $amt == 1;
		return 1;
	}
	else {
		return 0;
	}
}

sub get_flood_level($) {
	my ($user) = @_;

	$flood_check->execute(get_user_id($user));
	my ($level) = $flood_check->fetchrow_array;
	return $level;
}

sub get_uinfo(@) {
	my (@userlist) = @_;
	my @reply;
	foreach my $tuser (@userlist) {
		my ($ident, $vhost) = nickserv::get_vhost($tuser);
		my $modes = nickserv::get_user_modes($tuser);
		my $target = get_user_nick($tuser);
	
		my @data = (
			["Status:", (nickserv::is_online($tuser) ? "Online" : "Offline")],
			["ID Nicks:", join(', ', nickserv::get_id_nicks($tuser))],
			["Channels:", join(', ', chanserv::get_all_chans($tuser))],
			#push @reply,	"  Current server: ";
			["Flood level:", get_flood_level($tuser)],
			["Hostmask:", "$target\!$ident\@$vhost"],
			["Modes:", $modes]
		);
		push @data, ["Country:", country::get_country_long(country::get_user_country($tuser))]
			if module::is_loaded('country');
		push @reply, "User info for \002$target\002:", fmt::columnar(@data);
	}
	return @reply;
}

## IRC EVENTS ##

1;
